# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
from __future__ import annotations

from ryvencore import Node as NodeCore
from ryvencore.Base import Event

from ironflow.gui.canvas_widgets import NodeWidget


class Node(NodeCore):
    """
    A parent class for all ironflow nodes. Apart from a small quality-of-life difference where outputs are
    accessible in the same way as inputs (i.e. with a method `output(i)`), the main change here is the `before_update`
    and `after_update` events. Callbacks to happen before and after the update can be added to (removed from) these with
    the `connect` (`disconnect`) methods on the event. Such callbacks need to take the node itself as the first
    argument, and the integer specifying which input channel is being updated as the second argument.

    Also provides a "representation" that gets used in the GUI to give a more detailed look at node data, which defaults
    to showing output channel values.

    Children should specify a title and some combination of initial input, output, and what to do when updated, e.g.:

    >>> class My_Node(Node):
    >>> title = "MyUserNode"
    >>> init_inputs = [
    >>>     NodeInputBP(dtype=dtypes.Integer(default=1), label="foo")
    >>> ]
    >>> init_outputs = [
    >>>    NodeOutputBP(label="bar")
    >>> ]
    >>> color = 'cyan'
    >>>
    >>> def update_event(self, inp=-1):
    >>>     self.set_output_val(0, self.input(0) + 42)

    Note:
        When registering nodes from a module or .py file, only children of this class with names ending in `_Node` will
        get registered.
    """

    main_widget_class = NodeWidget

    color = "#ff69b4"  # Add an abrasive default color -- won't crash if you forget to add one, but pops out a bit

    def __init__(self, params):
        super().__init__(params)

        self.before_update = Event(self, int)
        self.after_update = Event(self, int)
        self.actions = (
            dict()
        )  # Resolves Todo from ryven.NENV, moving it to our node class instead of ryvencore

        self.representation_updated = False
        self.after_update.connect(self._representation_update)

    def place_event(self):
        # place_event() is executed *before* the connections are built
        super().place_event()
        for inp in self.inputs:
            if (
                inp.val is None
            ):  # Don't over-write data from loaded sessions with defaults
                if inp.dtype is not None:
                    inp.update(inp.dtype.default)
                elif "val" in inp.add_data.keys():
                    inp.update(inp.add_data["val"])

    def update(self, inp=-1):
        self.before_update.emit(self, inp)
        super().update(inp=inp)
        self.after_update.emit(self, inp)

    def output(self, i):
        return self.outputs[i].val

    @staticmethod
    def _representation_update(self, inp):
        self.representation_updated = True

    @property
    def representations(self) -> dict:
        return {
            o.label_str if o.label_str != "" else f"output{i}": o.val
            for i, o in enumerate(self.outputs)
            if o.type_ == "data"
        }


class PlaceholderWidgetsContainer:
    """
    An object that just returns None for all accessed attributes so widgets.MyWidget in the non-ironflow nodes files
    just returns None.
    """

    def __getattr__(self, item):
        return None



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Shared code among representations of node data.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

import ipywidgets as widgets

from ironflow.gui.boxes.base import Box


class NodeInterfaceBase(Box, ABC):
    box_class = widgets.Box

    def __init__(self):
        super().__init__()
        self.output = widgets.Output(layout={"width": "100%"})
        self.box.children = [self.output]

    @property
    def layout(self) -> widgets.Layout:
        return widgets.Layout(
            width="50%",
            border="1px solid black",
        )

    @abstractmethod
    def draw(self) -> None:
        pass

    def clear_output(self) -> None:
        self.output.clear_output()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Wrapped dtypes to avoid mutable defaults.

Implementation changes in ryvencore v0.4, so this file may be short-lived.
"""

from typing import Optional

from ryvencore.dtypes import DType, Data, Integer, Float, String, Boolean, Char


class Choice(DType):
    def __init__(
        self,
        default=None,
        items: Optional[list] = None,
        doc: str = "",
        _load_state=None,
    ):
        self.items = items if items is not None else []
        super().__init__(default=default, doc=doc, _load_state=_load_state)
        self.add_data("items")


class List(DType):
    def __init__(self, default: Optional[list] = None, doc: str = "", _load_state=None):
        default = default if default is not None else []
        super().__init__(default=default, doc=doc, _load_state=_load_state)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
For richer representations of node data allowing deeper analysis.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import ipywidgets as widgets
from IPython.display import display

from ironflow.gui.boxes.node_interface.base import NodeInterfaceBase

if TYPE_CHECKING:
    from ironflow.gui.canvas_widgets import NodeWidget


class NodePresenter(NodeInterfaceBase):
    """Handles the display of nodes with a representation."""

    def __init__(self):
        super().__init__()
        self._node_widget = None
        self._widgets = []
        self._toggles = []

    @property
    def node_widget(self) -> NodeWidget | None:
        return self._node_widget

    @node_widget.setter
    def node_widget(self, new_node_widget: NodeWidget | None):
        if self._node_widget is not None:
            self.clear_output()
            self._node_widget.represent_button.pressed = False
            self._node_widget.represent_button.draw()  # Re-draw it as un-pressed

        if new_node_widget is not None:
            new_node_widget.node.representation_updated = True
            self._node_widget = new_node_widget
            self._widgets = self._build_widgets(new_node_widget.node.representations)
            self._toggles = self._build_toggles(new_node_widget.node.representations)
        else:
            self._node_widget = None
            self._widgets = []
            self._toggles = []

    @staticmethod
    def _build_widgets(representations: dict) -> list[widgets.Output]:
        return [
            widgets.Output(layout={"border": "solid 1px gray"}) for _ in representations
        ]

    def _build_toggles(self, representations: dict) -> list[widgets.Checkbox]:
        toggles = []
        for i, label in enumerate(representations.keys()):
            toggle = widgets.Checkbox(
                description=label, value=i == 0, indent=False, layout={"width": "100px"}
            )
            toggle.observe(self._on_toggle)
            toggles.append(toggle)
        return toggles

    def _on_toggle(self, change: dict) -> None:
        if change["name"] == "value":
            self._draw()

    def _draw(self):
        self.clear_output()

        representations = []
        for (toggle, widget, representation) in zip(
            self._toggles, self._widgets, self.node_widget.node.representations.values()
        ):
            if toggle.value:
                with widget:
                    display(representation)
                representations.append(widget)

        with self.output:
            display(
                widgets.VBox(
                    [
                        widgets.HBox(self._toggles, layout={"flex_flow": "row wrap"}),
                        *representations,
                    ]
                )
            )

    def draw(self) -> None:
        if (
            self.node_widget is not None
            and self.node_widget.node.representation_updated
        ):
            self._draw()
            self.node_widget.node.representation_updated = False

    def clear_output(self) -> None:
        for w in self._widgets:
            w.clear_output()
        super().clear_output()

    def close(self) -> None:
        self.node_widget = None
        self.clear_output()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
For directly controlling node IO (as opposed to giving it to nodes by connecting them to other OI).
"""

from __future__ import annotations

import pickle
import base64
from typing import TYPE_CHECKING, Callable

import ipywidgets as widgets
import numpy as np
from IPython.display import display

from ironflow.gui.boxes.node_interface.base import NodeInterfaceBase


if TYPE_CHECKING:
    from ironflow.gui.gui import GUI
    from ironflow.model.node import Node


def deserialize(data):
    return pickle.loads(base64.b64decode(data))


class NodeController(NodeInterfaceBase):
    """
    Handles the creation of widgets for manually adjusting node input and viewing node info.
    """

    def __init__(self, gui: GUI):
        super().__init__()
        self.gui = gui
        self.node = None
        self._margin = 5  # px
        self._row_height = 30  # px

    def _box_height(self, n_rows: int) -> int:
        return n_rows * self._row_height + 2 * self._margin

    @property
    def input_widget(self) -> widgets.Widget:
        try:
            widget = self.node.input_widget(self.gui, self.node).widget
            widget.layout = widgets.Layout(
                height="70px",
                border="solid 1px blue",
                margin=f"{self._margin}px",
                padding="10px",
                width="auto",
            )
            return widget
        except AttributeError:
            return widgets.Output()

    def input_field_list(self) -> list[list[widgets.Widget]]:
        input = []
        if hasattr(self.node, "inputs"):
            for i_c, inp in enumerate(self.node.inputs[:]):
                if inp.dtype is not None:
                    dtype = str(inp.dtype).split(".")[-1]
                    dtype_state = deserialize(inp.data()["dtype state"])
                    if inp.val is None:
                        inp.val = dtype_state["val"]
                    if dtype == "Integer":
                        inp_widget = widgets.IntText(
                            value=inp.val,
                            disabled=False,
                            description="",
                            continuous_update=False,
                        )
                    elif dtype == "Boolean":
                        inp_widget = widgets.Checkbox(
                            value=inp.val,
                            indent=False,
                            description="",
                        )
                    elif dtype == "Choice":
                        inp_widget = widgets.Dropdown(
                            value=inp.val,
                            options=inp.dtype.items,
                            description="",
                            ensure_option=True,
                        )

                    else:
                        inp_widget = widgets.Text(
                            value=str(inp.val),
                            continuous_update=False,
                        )
                    description = inp.label_str
                elif inp.label_str != "":
                    inp_widget = widgets.Label(value=inp.type_)
                    description = inp.label_str
                else:
                    inp_widget = widgets.Label(value=inp.type_)
                    description = inp.type_
                inp_widget.observe(self.input_change_i(i_c), names="value")
                input.append([widgets.Label(description), inp_widget])
        return input

    def input_change_i(self, i_c) -> Callable:
        def input_change(change: dict) -> None:
            # Todo: Test this in exec mode
            self.node.inputs[i_c].val = change["new"]
            self.node.update(i_c)
            self.gui.redraw_active_flow_canvas()

        return input_change

    @property
    def input_box(self) -> widgets.GridBox | widgets.Output:
        input_fields = self.input_field_list()
        n_fields = len(input_fields)
        if n_fields > 0:
            return widgets.GridBox(
                list(np.array(input_fields).flatten()),
                layout=widgets.Layout(
                    grid_template_columns="110px auto",
                    grid_auto_rows=f"{self._row_height}px",
                    border="solid 1px blue",
                    margin=f"{self._margin}px",
                    height=f"{self._box_height(n_fields)}px",
                    # Automatic height like this really should be doable just with the CSS,
                    # but for the life of me I can't get a CSS solution working right -Liam
                ),
            )
        else:
            return widgets.Output()

    @property
    def info_box(self) -> widgets.VBox:
        glob_id_val = None
        if hasattr(self.node, "GLOBAL_ID"):
            glob_id_val = self.node.GLOBAL_ID
        global_id = widgets.Text(
            value=str(glob_id_val), description="GLOBAL_ID:", disabled=True
        )

        title = widgets.Text(
            value=str(self.node.title), description="Title:", disabled=True
        )

        info_box = widgets.VBox([title, global_id])
        info_box.layout = widgets.Layout(
            height=f"{self._box_height(2)}px",
            border="solid 1px red",
            margin=f"{self._margin}px",
            padding="0px",
        )
        return info_box

    def draw(self) -> None:
        self.clear_output()
        if self.node is not None:
            with self.output:
                display(
                    widgets.VBox([self.input_box, self.input_widget, self.info_box])
                )
                # PyCharm nit is invalid, display takes *args is why it claims to want a tuple

    def draw_for_node(self, node: Node | None) -> None:
        self.node = node
        self.draw()

    def close(self) -> None:
        self.draw_for_node(None)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Ryven nodes specifc to pyiron (or with ironflow improvements like an ipywidgets representation).
"""

from __future__ import annotations

import json
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import matplotlib.pylab as plt
import numpy as np

from pyiron_atomistics import Project
from pyiron_atomistics.atomistics.structure.factory import StructureFactory
from pyiron_atomistics.atomistics.job.atomistic import AtomisticGenericJob
from pyiron_atomistics.lammps import list_potentials

from ironflow.gui.canvas_widgets import ButtonNodeWidget
from ironflow.model import dtypes, NodeInputBP, NodeOutputBP
from ironflow.model.node import Node
from ironflow.nodes.std.special_nodes import DualNodeBase

if TYPE_CHECKING:
    from pyiron_base import HasGroups

STRUCTURE_FACTORY = StructureFactory()


class BeautifulHasGroups:
    """
    A helper class for giving classes that inherit from `pyiron_base.HasGroups` a more appealing representation in
    ipywidgets.
    """

    def __init__(self, has_groups: HasGroups | None):
        self._has_groups = has_groups

    def to_builtin(self, has_groups=None):
        has_groups = has_groups if has_groups is not None else self._has_groups
        if has_groups is not None:
            repr_dict = {}
            for k in has_groups.list_groups():
                repr_dict[k] = self.to_builtin(has_groups[k])
            for k in has_groups.list_nodes():
                repr_dict[k] = str(has_groups[k])
            return repr_dict
        else:
            return None

    def _repr_json_(self):
        return self.to_builtin()

    def _repr_html_(self):
        name = self._has_groups.__class__.__name__
        plain = f"{name}({json.dumps(self.to_builtin(), indent=2, default=str)})"
        return "<pre>" + plain + "</pre>"


class Project_Node(Node):
    """
    Create a pyiron project.

    Inputs:
        name (str): The name of the project. Will access existing project data under that name. (Default is ".".)

    Outputs:
        project (pyiron_atomistics.Project): The project object.
    """

    # this __doc__ string will be displayed as tooltip in the editor

    title = "Project"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Char(default="."), label="name"),
    ]
    init_outputs = [
        NodeOutputBP(label="project"),
    ]
    color = "#aabb44"

    def place_event(self):
        super().place_event()
        self.update()

    def update_event(self, inp=-1):
        pr = Project(self.input(0))
        self.set_output_val(0, pr)

    @property
    def _project(self):
        return self.output(0)

    @property
    def representations(self) -> dict:
        return {
            "name": str(self.input(0)),
            # "job_table": self._project.job_table() if self._project is not None else None
            # Todo: Figure out how to display this without breaking the gui size; right now it automatically grows
            # the gui because the table is so wide.
        }


class OutputsOnlyAtoms(Node, ABC):
    """
    A helper class that manages representations for nodes whose only output is a `pyiron_atomistics.Atoms` object.

    Outputs:
        structure (pyiron_atomistics.Atoms): An atomic structure.
    """

    init_outputs = [
        NodeOutputBP(label="structure"),
    ]
    color = "#aabb44"

    @abstractmethod
    def update_event(self, inp=-1):
        """Must set output 0 to an instance of pyiron_atomistics.atomistics.atoms.Atoms"""
        pass

    @property
    def representations(self) -> dict:
        return {"plot3d": self.output(0).plot3d(), "print": self.output(0)}


class BulkStructure_Node(OutputsOnlyAtoms):
    """
    Generate a bulk atomic structure.

    Inputs:
        element (str): The atomic symbol for the desired atoms. (Default is "Fe".)
        crystal_structure (str | None): Must be one of sc, fcc, bcc, hcp, diamond, zincblende,
                                rocksalt, cesiumchloride, fluorite or wurtzite.
        a (float | None): Lattice constant.
        c (float | None): Lattice constant.
        c_over_a (float | None): c/a ratio used for hcp.  Default is ideal ratio: sqrt(8/3).
        u (float | None): Internal coordinate for Wurtzite structure.
        orthorhombic (bool): Construct orthorhombic unit cell instead of primitive cell. (Takes precedence over cubic
            flag when both are true.)
        cubic (bool): Construct cubic unit cell if possible.

    Outputs:
        structure (pyiron_atomistics.Atoms): A mono-species bulk structure.
    """

    # this __doc__ string will be displayed as tooltip in the editor

    title = "BulkStructure"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Char(default="Fe"), label="element"),
        NodeInputBP(
            dtype=dtypes.Choice(
                default=None,
                items=[
                    None,
                    "sc",
                    "fcc",
                    "bcc",
                    "hcp",
                    "diamond",
                    "zincblende",
                    "rocksalt",
                    "cesiumchloride",
                    "fluorite",
                    "wurtzite",
                ],
            ),
            label="crystal_structure",
        ),
        NodeInputBP(dtype=dtypes.Float(default=None), label="a"),
        NodeInputBP(dtype=dtypes.Float(default=None), label="c"),
        NodeInputBP(dtype=dtypes.Float(default=None), label="c_over_a"),
        NodeInputBP(dtype=dtypes.Float(default=None), label="u"),
        NodeInputBP(dtype=dtypes.Boolean(default=False), label="orthorhombic"),
        NodeInputBP(dtype=dtypes.Boolean(default=False), label="cubic"),
    ]

    def update_event(self, inp=-1):
        self.set_output_val(
            0,
            STRUCTURE_FACTORY.bulk(
                self.input(0),
                crystalstructure=self.input(1),
                a=self.input(2),
                c=self.input(3),
                covera=self.input(4),
                u=self.input(5),
                orthorhombic=self.input(6),
                cubic=self.input(7),
            ),
        )

    def place_event(self):
        super().place_event()
        self.update()


class Repeat_Node(OutputsOnlyAtoms):
    """
    Repeat atomic structure supercell.

    Inputs:
        structure (pyiron_atomistics.Atoms): The structure to repeat periodically.
        all (int): The number of times to repeat it in each of the three bravais lattice directions.

    Outputs:
        structure (pyiron_atomistics.Atoms): A repeated copy of the input structure.
    """

    # this __doc__ string will be displayed as tooltip in the editor

    title = "Repeat"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Data(size="m"), label="structure"),
        NodeInputBP(dtype=dtypes.Integer(default=1, bounds=(1, 100)), label="all"),
    ]

    def update_event(self, inp=-1):
        self.set_output_val(0, self.input(0).repeat(self.input(1)))


class ApplyStrain_Node(OutputsOnlyAtoms):
    """
    Apply strain on atomic structure supercell.

    Inputs:
        structure (pyiron_atomistics.Atoms): The atomic structure to strain.
        strain (float): The isotropic strain to apply, where 0 is unstrained. (Default is 0.)

    Outputs:
        (pyiron_atomistics.Atoms): A strained copy of the input structure.
    """

    title = "ApplyStrain"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Data(size="m"), label="structure"),
        NodeInputBP(dtype=dtypes.Float(default=0, bounds=(-100, 100)), label="strain"),
    ]

    def update_event(self, inp=-1):
        self.set_output_val(
            0, self.input(0).apply_strain(float(self.input(1)), return_box=True)
        )


class Lammps_Node(Node):
    """
    WIP.
    """

    title = "Lammps"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec", label="run"),
        NodeInputBP(type_="exec", label="remove"),
        NodeInputBP(dtype=dtypes.Data(size="m"), label="project"),
        NodeInputBP(dtype=dtypes.Char(default="job"), label="name"),
        NodeInputBP(dtype=dtypes.Data(size="m"), label="structure"),
        NodeInputBP(
            dtype=dtypes.Choice(
                default="Set structure first", items=["Set structure first"]
            ),
            label="potential",
        ),
    ]
    init_outputs = [
        NodeOutputBP(type_="exec"),
        NodeOutputBP(label="job"),
    ]
    color = "#5d95de"

    @property
    def _project(self):
        return self.input(2)

    @property
    def _name(self):
        return self.input(3)

    @property
    def _structure(self):
        return self.input(4)

    @property
    def _potential(self):
        return self.input(5)

    def _run(self):
        job = self._project.create.job.Lammps(self._name)
        job.structure = self._structure
        job.potential = self._potential
        self._job = job
        job.run()
        self.set_output_val(1, job)
        self.exec_output(0)

    def _remove(self):
        try:
            name = (
                self._job.name
            )  # Remove based on the run job, not the input name which might have changed...
            self._project.remove_job(name)
            self.set_output_val(1, None)
        except AttributeError:
            pass

    def _update_potential_choices(self):
        potl_input = self.inputs[5]
        last_potential = potl_input.val
        structure = self.inputs[4].val
        available_potentials = list_potentials(structure)

        if len(available_potentials) == 0:
            potl_input.val = "No valid potential"
            potl_input.dtype.items = ["No valid potential"]
        else:
            if last_potential not in available_potentials:
                potl_input.val = available_potentials[0]
            potl_input.dtype.items = available_potentials

    def update_event(self, inp=-1):
        if inp == 0:
            self._run()
        elif inp == 1:
            self._remove()
        elif inp == 4:
            self._update_potential_choices()

    @property
    def representations(self) -> dict:
        return {"job": BeautifulHasGroups(self.output(1))}


class GenericOutput_Node(Node):
    """
    Select Generic Output item.

    Inputs:
        job (AtomisticGenericJob): A job with an `output` attribute of type
            `pyiron_atomistics.atomistics.job.atomistic.GenericOutput`.
        field (dtypes.Choice): Which output field to look at. Automatically populates once the job is valid.

    Outputs:
        output (numpy.ndarray): The selected output field.
    """

    version = "v0.1"
    title = "GenericOutput"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Data(size="m"), label="job"),
        NodeInputBP(
            dtype=dtypes.Choice(
                default="Input an atomistic job", items=["Input an atomistic job"]
            ),
            label="field",
        ),
    ]
    init_outputs = [
        NodeOutputBP(label="output"),
    ]
    color = "#c69a15"

    def __init__(self, params):
        super().__init__(params)

    @property
    def _job(self):
        return self.input(0)

    def _update_fields(self):
        if isinstance(self._job, AtomisticGenericJob):
            self.inputs[1].dtype.items = self._job["output/generic"].list_nodes()
            self.inputs[1].val = self.inputs[1].dtype.items[0]
        else:
            self.inputs[1].dtype.items = [self.init_inputs[1].dtype.default]
            # Note: It would be sensible to use `self.init_outputs[1].dtype.items` above, but this field gets updated
            # to `self.inputs[1].dtype.items`, probably because of the mutability of lists.
            self.inputs[1].val = self.init_inputs[1].dtype.default

    def _update_value(self):
        if isinstance(self._job, AtomisticGenericJob):
            val = self._job[f"output/generic/{self.input(1)}"]
        else:
            val = None
        self.set_output_val(0, val)

    def update_event(self, inp=-1):
        if inp == 0:
            self._update_fields()
            self._update_value()
        elif inp == 1:
            self._update_value()


class IntRand_Node(Node):
    """
    Generate a random non-negative integer.

    Inputs:
        high (int): Biggest possible integer. (Default is 1).
        length (int): How many random numbers to generate. (Default is 1.)

    Outputs:
        randint (int|numpy.ndarray): The randomly generated value(s).
    """

    # this __doc__ string will be displayed as tooltip in the editor

    title = "IntRandom"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Integer(default=1, bounds=(10, 100)), label="high"),
        NodeInputBP(dtype=dtypes.Integer(default=1, bounds=(1, 100)), label="length"),
    ]
    init_outputs = [
        NodeOutputBP(label="randint"),
    ]
    color = "#aabb44"

    def update_event(self, inp=-1):
        val = np.random.randint(0, high=self.input(0), size=self.input(1))
        self.set_output_val(0, val)


class JobName_Node(Node):
    """
    Create job name for parameters.

    Inputs:
        base (str): The stem for the final name. (Default is "job_".)
        float (float): The parameter value to add to the name.

    Outputs:
        job_name (str): The base plus float sanitized into a valid job name.

    Todo:
        There has been some work in pyiron_base on getting a cleaner job name sanitizer, so lean on that.
    """

    title = "JobName"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Char(default="job_"), label="base"),
        NodeInputBP(dtype=dtypes.Float(default=0), label="float"),
    ]
    init_outputs = [
        NodeOutputBP(label="job_name"),
    ]
    color = "#aabb44"

    def update_event(self, inp=-1):
        val = self.input(0) + f"{float(self.input(1))}".replace("-", "m").replace(
            ".", "p"
        )
        self.set_output_val(0, val)


class Linspace_Node(Node):
    """
    Generate a linear mesh in a given range using `np.linspace`.

    Inputs:
        min (int): The lower bound (inclusive). (Default is 1.)
        max (int): The upper bound (inclusive). (Default is 2.)
        steps (int): How many samples to take inside (min, max). (Default is 10.)

    Outputs:
        linspace (numpy.ndarray): A uniform sampling over the requested range.
    """

    # this __doc__ string will be displayed as tooltip in the editor

    title = "Linspace"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Integer(default=1, bounds=(0, 100)), label="min"),
        NodeInputBP(dtype=dtypes.Integer(default=2, bounds=(0, 100)), label="max"),
        NodeInputBP(dtype=dtypes.Integer(default=10, bounds=(1, 100)), label="steps"),
    ]
    init_outputs = [
        NodeOutputBP(label="linspace"),
    ]
    color = "#aabb44"

    def place_event(self):
        super().place_event()
        self.update()

    def update_event(self, inp=-1):
        val = np.linspace(self.input(0), self.input(1), self.input(2))
        self.set_output_val(0, val)


class Plot3d_Node(Node):
    """
    Plot a structure with NGLView.

    Inputs:
        structure (pyiron_atomistics.Atoms): The structure to plot.

    Outputs:
        plot3d (nglview.widget.NGLWidget): The plot object.
        structure (pyiron_atomistics.Atoms): The raw structure object passed in.
    """

    title = "Plot3d"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Data(size="m"), label="structure"),
    ]
    init_outputs = [
        NodeOutputBP(type_="data", label="plot3d"),
        NodeOutputBP(type_="data", label="structure"),
    ]
    color = "#5d95de"

    def update_event(self, inp=-1):
        self.set_output_val(0, self.input(0).plot3d())
        self.set_output_val(1, self.input(0))


class Matplot_Node(Node):
    """
    A 2D matplotlib plot.

    Inputs:
        x (list|numpy.ndarray|...): Data for the x-axis.
        y (list|numpy.ndarray|...): Data for the y-axis.

    Outputs:
        fig (matplotlib.figure.Figure): The resulting figure after a `matplotlib.pyplot.plot` call on x and y.
    """

    title = "MatPlot"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Data(size="m"), label="x"),
        NodeInputBP(dtype=dtypes.Data(size="m"), label="y"),
    ]
    init_outputs = [
        NodeOutputBP(type_="data", label="fig"),
    ]
    color = "#5d95de"

    def update_event(self, inp=-1):
        super().update_event()
        plt.ioff()
        fig = plt.figure()
        plt.clf()
        plt.plot(self.input(0), self.input(1))
        self.set_output_val(0, fig)
        plt.ion()


class Sin_Node(Node):
    """
    Call `numpy.sin` on a value.

    Inputs:
        x (int|float|list|numpy.ndarray|...): The value to sine transform.

    Outputs:
        sin (float|numpy.ndarray): The sine of x.
    """

    title = "Sin"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Data(size="m"), label="x"),
    ]
    init_outputs = [
        NodeOutputBP(label="sin"),
    ]
    color = "#5d95de"

    def update_event(self, inp=-1):
        self.set_output_val(0, np.sin(self.input(0)))


class Result_Node(Node):
    """Simply shows a value converted to str"""

    version = "v0.1"

    title = "Result"
    init_inputs = [
        NodeInputBP(type_="data"),
    ]
    color = "#c69a15"

    def __init__(self, params):
        super().__init__(params)
        self.val = None

    def place_event(self):
        super().place_event()
        self.update()

    def view_place_event(self):
        self.main_widget().show_val(self.val)

    def update_event(self, inp=-1):
        self.val = self.input(0)
        if self.session.gui:
            self.main_widget().show_val(self.val)


class ForEach_Node(Node):
    title = "ForEach"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec", label="start"),
        NodeInputBP(type_="exec", label="reset"),
        NodeInputBP(dtype=dtypes.List(), label="elements"),
    ]
    init_outputs = [
        NodeOutputBP(label="loop", type_="exec"),
        NodeOutputBP(label="e", type_="data"),
        NodeOutputBP(label="finished", type_="exec"),
    ]
    color = "#b33a27"

    _count = 0

    def update_event(self, inp=-1):
        if inp == 0:
            self._count += 1
            if len(self.input(2)) > self._count:
                e = self.input(2)[self._count]
                self.set_output_val(1, e)
                self.exec_output(0)
            else:
                self.exec_output(2)
        elif inp > 0:
            self._count = 0
        self.val = self._count


class ExecCounter_Node(DualNodeBase):
    title = "ExecCounter"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec"),
    ]
    init_outputs = [
        NodeOutputBP(type_="exec"),
    ]
    color = "#5d95de"

    def __init__(self, params):
        super().__init__(params, active=True)
        self._count = 0

    def update_event(self, inp=-1):
        if self.active and inp == 0:
            self._count += 1
            self.val = self._count
        elif not self.active:
            self.val = self.input(0)


class Click_Node(Node):
    title = "Click"
    version = "v0.1"
    main_widget_class = ButtonNodeWidget
    init_inputs = []
    init_outputs = [NodeOutputBP(type_="exec")]
    color = "#99dd55"

    def update_event(self, inp=-1):
        self.exec_output(0)


nodes = [
    Project_Node,
    BulkStructure_Node,
    Repeat_Node,
    ApplyStrain_Node,
    Lammps_Node,
    JobName_Node,
    GenericOutput_Node,
    Plot3d_Node,
    IntRand_Node,
    Linspace_Node,
    Sin_Node,
    Result_Node,
    ExecCounter_Node,
    Matplot_Node,
    Click_Node,
    ForEach_Node,
]



from ironflow.model import dtypes, NodeInputBP, NodeOutputBP
from ironflow.model.node import Node


class CSNodeBase(Node):
    version = "v0.1"
    style = "normal"
    color = "#b33a27"


class If_Node(CSNodeBase):
    title = "branch"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec"),
        NodeInputBP(dtype=dtypes.Boolean(), label="cond"),
    ]
    init_outputs = [
        NodeOutputBP(
            "true",
            type_="exec",
        ),
        NodeOutputBP("false", type_="exec"),
    ]

    def update_event(self, inp=-1):
        if inp == 0:
            if self.input(1):
                self.exec_output(0)
            else:
                self.exec_output(1)


class ForLoop_Node(CSNodeBase):
    """n dimensional for loop!"""

    title = "for"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec"),
        NodeInputBP(dtype=dtypes.Integer(), label="from"),
        NodeInputBP(dtype=dtypes.Integer(), label="to"),
    ]
    init_outputs = [
        NodeOutputBP("loop", type_="exec"),
        NodeOutputBP("i", type_="data"),
        NodeOutputBP("finished", type_="exec"),
    ]

    def __init__(self, params):
        super().__init__(params)

        self.actions["add dimension"] = {"method": self.add_dimension}

        self.dims = 1

    def add_dimension(self):
        new_dim = self.dims + 1
        self.create_input_dt(dtype=dtypes.Integer(), label=f"{new_dim} from")
        self.create_input_dt(dtype=dtypes.Integer(), label=f"{new_dim} to")
        self.create_output(f"loop {new_dim}", "exec", insert=-1)
        self.create_output(f"i {new_dim}", "data", insert=-1)
        self.dims += 1

        self.actions[f"remove dimension {new_dim}"] = {
            "method": self.remove_dimension,
            "data": new_dim,
        }

    def remove_dimension(self, dim):
        inp_index = self.input_from_dim(dim)
        self.delete_input(inp_index)
        self.delete_input(inp_index)
        out_index = self.output_from_dim(dim)
        self.delete_output(out_index)
        self.delete_output(out_index)
        self.dims -= 1
        # del self.actions[f'remove dimension {dim}']
        self.rebuild_remove_actions()

    def rebuild_remove_actions(self):

        remove_keys = []
        for k, v in self.actions.items():
            if k.startswith("remove dimension"):
                remove_keys.append(k)

        for k in remove_keys:
            del self.actions[k]

        for i in range(1, self.dims):
            self.actions[f"remove dimension {i+1}"] = {
                "method": self.remove_dimension,
                "data": i + 1,
            }

    def input_from_dim(self, dim):
        return 1 + 2 * (dim - 1)

    def output_from_dim(self, dim):
        return 2 * (dim - 1)

    def update_event(self, inp=-1):
        if inp == 0:
            self.iterate(1)
            self.exec_output(len(self.outputs) - 1)

    def iterate(self, dim):

        inp_index = self.input_from_dim(dim)

        exec_out_index = self.output_from_dim(dim)
        data_out_index = exec_out_index + 1

        for i in range(self.input(inp_index), self.input(inp_index + 1)):

            self.set_output_val(data_out_index, i)
            self.exec_output(exec_out_index)

            if dim < self.dims:
                self.iterate(dim + 1)


class ForEachLoop_Node(CSNodeBase):
    title = "for each"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec"),
        NodeInputBP(dtype=dtypes.Integer(), label="elements"),
    ]
    init_outputs = [
        NodeOutputBP("loop", type_="exec"),
        NodeOutputBP("e", type_="data"),
        NodeOutputBP("finished", type_="exec"),
    ]

    def update_event(self, inp=-1):
        for e in self.input(0):
            self.set_output_val(1, e)
            self.exec_output(0)

        self.exec_output(2)


class WhileLoop_Node(CSNodeBase):
    title = "while"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec"),
        NodeInputBP(dtype=dtypes.Boolean(), label="cond"),
    ]
    init_outputs = [
        NodeOutputBP("loop", type_="exec"),
        NodeOutputBP("finished", type_="exec"),
    ]

    def update_event(self, inp=-1):
        while self.input(0):
            self.exec_output(0)

        self.exec_output(1)


class DoWhileLoop_Node(CSNodeBase):
    title = "do while"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec"),
        NodeInputBP(dtype=dtypes.Boolean(), label="cond"),
    ]
    init_outputs = [
        NodeOutputBP("loop", type_="exec"),
        NodeOutputBP("finished", type_="exec"),
    ]

    def update_event(self, inp=-1):
        self.exec_output(0)
        while self.input(0):
            self.exec_output(0)
        self.exec_output(1)


nodes = [
    If_Node,
    ForLoop_Node,
    ForEachLoop_Node,
    WhileLoop_Node,
    DoWhileLoop_Node,
]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
For getting non-flow feedback from the user, e.g. filenames and confirmations.
"""

from __future__ import annotations

from typing import Any, Optional

import ipywidgets as widgets

from ironflow.gui.boxes.base import Box


class UserInput(Box):
    box_class = widgets.HBox

    def __init__(self):
        super().__init__()
        self.input_field = widgets.Text(value="INIT VALUE", description="DESCRIPTION")
        self.decision_info = widgets.Label(value="INIT VALUE")
        button_layout = widgets.Layout(width="50px")
        self.ok_button = widgets.Button(
            tooltip="Confirm", icon="check", layout=button_layout
        )
        self._last_ok_callback = None
        self.cancel_button = widgets.Button(
            tooltip="Cancel", icon="ban", layout=button_layout
        )
        # Todo: Use xmark once this is available
        self.cancel_button.on_click(self.close)

    def clear(self):
        super().clear()
        self._clear_callback()

    @property
    def text(self):
        return self.input_field.value

    def wrap_callback(self, callback: callable) -> callable:
        def wrapped_callback(change: dict) -> None:
            callback(change)
            self.clear()

        return wrapped_callback

    def _clear_callback(self):
        if self._last_ok_callback is not None:
            self.input_field.on_submit(self._last_ok_callback, remove=True)
            self.ok_button.on_click(self._last_ok_callback, remove=True)

    def _set_callback(self, callback: callable):
        wrapped_callback = self.wrap_callback(callback)
        self.input_field.on_submit(wrapped_callback)
        # ^ Ignore the deprecation warning, 'observe' doesn't function the way we actually want
        # https://github.com/jupyter-widgets/ipywidgets/issues/2446
        self.ok_button.on_click(wrapped_callback)
        self._last_ok_callback = wrapped_callback

    def _open(
        self,
        widget: widgets.Widget,
        callback: callable,
        ok_tooltip: str,
        cancel_tooltip: str,
    ):
        self._box.children = [widget, self.ok_button, self.cancel_button]
        self.ok_button.tooltip = ok_tooltip
        self.cancel_button.tooltip = cancel_tooltip
        self._set_callback(callback)

    def open_text(
        self,
        description: str,
        callback: callable,
        initial_value: Any,
        description_tooltip: Optional[str] = None,
        ok_tooltip: str = "Confirm",
        cancel_tooltip: str = "Cancel",
    ):
        self.input_field.description = description
        description_tooltip = (
            description_tooltip if description_tooltip is not None else description
        )
        self.input_field.description_tooltip = description_tooltip
        self.input_field.value = initial_value
        self._open(self.input_field, callback, ok_tooltip, cancel_tooltip)

    def open_bool(
        self,
        description: str,
        callback: callable,
        ok_tooltip: str = "Confirm",
        cancel_tooltip: str = "Cancel",
    ):
        self.decision_info.value = description
        self._open(self.decision_info, callback, ok_tooltip, cancel_tooltip)

    def close(self, change: None):
        self.clear()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Wrappers and extensions of the underlying ryven model.
"""

from ryvencore import (
    Flow,
    NodeInputBP,
    NodeOutputBP,
)
from ryvencore.NodePort import NodePort, NodeInput, NodeOutput



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Component widgets from ipywidgets that get assembled to make the top-level gui.
"""

from ironflow.gui.boxes.node_interface import NodeController, NodePresenter
from ironflow.gui.boxes.text_output import TextOut
from ironflow.gui.boxes.user_input import UserInput
from ironflow.gui.boxes.toolbar import Toolbar
from ironflow.gui.boxes.flow import FlowBox



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
For giving text feedback to the user.
"""

from __future__ import annotations

import ipywidgets as widgets
from IPython.display import display

from ironflow.gui.boxes.base import Box


class TextOut(Box):
    box_class = widgets.VBox

    def __init__(self):
        super().__init__()
        self._output = widgets.Output()
        self._button = widgets.Button(
            tooltip="Clear output",
            description="clear",
            layout={"width": "100px"},
        )
        self._button.on_click(self._click_button)

    @property
    def layout(self):
        return widgets.Layout(
            width="100%",
            border="1px solid black",
        )

    def clear(self):
        super().clear()
        self._output.clear_output()

    def _click_button(self, change: dict) -> None:
        self.clear()

    def print(self, msg: str):
        self._output.clear_output()
        with self._output:
            display(msg)
        self.box.children = [self._output, self._button]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Shared code among various canvas widgets.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Union

from ipycanvas import Canvas

from ironflow.gui.canvas_widgets.layouts import Layout

Number = Union[int, float]
if TYPE_CHECKING:
    from ironflow.gui.gui import GUI
    from ironflow.gui.canvas_widgets.flow import FlowCanvas
    from ironflow.model import Flow


class CanvasWidget(ABC):
    """
    Parent class for all "widgets" that exist inside the scope of the flow canvas.
    """

    def __init__(
        self,
        x: Number,
        y: Number,
        parent: FlowCanvas | CanvasWidget,
        layout: Layout,
        selected: bool = False,
        title: Optional[str] = None,
    ):
        self._x = x  # relative to parent
        self._y = y

        self.layout = layout
        self.parent = parent
        self._selected = selected
        self.title = title

        self.objects_to_draw = []

        self._height = self.layout.height

    @abstractmethod
    def on_click(
        self, last_selected_object: Optional[CanvasWidget]
    ) -> CanvasWidget | None:
        pass

    def on_double_click(self) -> CanvasWidget | None:
        return self

    @property
    def width(self) -> int:
        return self.layout.width

    @property
    def height(self) -> int:
        return self._height

    @property
    def x(self) -> Number:
        return self.parent.x + self._x

    @property
    def y(self) -> Number:
        return self.parent.y + self._y

    @property
    def canvas(self) -> Canvas:
        return self.parent.canvas

    @property
    def gui(self) -> GUI:
        return self.parent.gui

    @property
    def flow(self) -> Flow:
        return self.parent.flow

    @property
    def flow_canvas(self) -> FlowCanvas:
        return self.parent.flow_canvas

    def deselect_all(self) -> None:
        return self.parent.deselect_all()

    def add_widget(self, widget: CanvasWidget) -> None:
        self.objects_to_draw.append(widget)

    def set_x_y(self, x_in: Number, y_in: Number) -> None:
        self._x = x_in - self.width // 2
        self._y = y_in - self.height // 2

    def add_x_y(self, dx_in: Number, dy_in: Number) -> None:
        self._x += dx_in
        self._y += dy_in

    @property
    def color(self) -> str:
        return (
            self.layout.selected_color
            if self.selected
            else self.layout.background_color
        )

    def draw_shape(self) -> None:
        self.canvas.fill_style = self.color
        self.canvas.fill_rect(
            self.x,
            self.y,
            self.width,
            self.height,
        )

    @abstractmethod
    def draw_title(self) -> None:
        pass

    def draw(self) -> None:
        self.draw_shape()
        if self.title is not None:
            self.draw_title()
        for o in self.objects_to_draw:
            o.draw()

    def _is_at_xy(self, x_in: Number, y_in: Number) -> bool:
        x_coord = self.x
        y_coord = self.y
        return x_coord < x_in < (x_coord + self.width) and y_coord < y_in < (
            y_coord + self.height
        )

    def get_element_at_xy(self, x_in: Number, y_in: Number) -> CanvasWidget | None:
        if self.is_here(x_in, y_in):
            for o in self.objects_to_draw:
                if o.is_here(x_in, y_in):
                    return o.get_element_at_xy(x_in, y_in)
            return self
        else:
            return None

    def is_here(self, x_in: Number, y_in: Number) -> bool:
        return self._is_at_xy(x_in, y_in)

    def select(self) -> None:
        self._selected = True

    def deselect(self) -> None:
        self._selected = False
        [o.deselect() for o in self.objects_to_draw]

    @property
    def selected(self):
        return self._selected


class HideableWidget(CanvasWidget, ABC):
    def __init__(
        self,
        x: Number,
        y: Number,
        parent: FlowCanvas | CanvasWidget,
        layout: Layout,
        selected: bool = False,
        title: Optional[str] = None,
        visible: bool = True,
        hidden_x: Optional[Number] = None,
        hidden_y: Optional[Number] = None,
    ):
        super().__init__(
            x=x, y=y, parent=parent, layout=layout, selected=selected, title=title
        )
        self._hidden_x = hidden_x if hidden_x is not None else x
        self._hidden_y = hidden_y if hidden_y is not None else y
        self.visible = visible

    @property
    def hidden_x(self) -> Number:
        return self.parent.x + self._hidden_x

    @property
    def hidden_y(self) -> Number:
        return self.parent.y + self._hidden_y

    @property
    def x(self) -> Number:
        if self.visible:
            return self.parent.x + self._x
        else:
            return self.hidden_x

    @property
    def y(self) -> Number:
        if self.visible:
            return self.parent.y + self._y
        else:
            return self.hidden_y

    def show(self):
        self.visible = True

    def hide(self):
        self.visible = False

    def is_here(self, x_in: Number, y_in: Number) -> bool:
        if self.visible:
            return self._is_at_xy(x_in=x_in, y_in=y_in)
        else:
            return False

    def draw(self) -> None:
        if self.visible:
            super().draw()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
More detailed representations of data from the nodes.
"""

from ironflow.gui.boxes.node_interface.control import NodeController
from ironflow.gui.boxes.node_interface.representation import NodePresenter



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
A place to collect all the UI buttons.
"""

from __future__ import annotations

import ipywidgets as widgets

from ironflow.gui.boxes.base import Box


class Buttons:
    def __init__(self):
        """Toolbar buttons, declared in the order they will appear."""
        layout = widgets.Layout(width="50px")
        # Icon source: https://fontawesome.com
        # It looks like I'm stuck on v4, but this might just be a limitation of my jupyter environment -Liam
        # v4 icon search: https://fontawesome.com/v4/icons/
        self.load = widgets.Button(
            tooltip="Load session from JSON", icon="upload", layout=layout
        )
        self.save = widgets.Button(
            tooltip="Save session to JSON", icon="download", layout=layout
        )
        self.help_node = widgets.Button(
            tooltip="Print docs for new node class",
            icon="question-circle",
            layout=layout,
        )
        self.add_node = widgets.Button(
            tooltip="Add new node (or double-click on empty space)",
            icon="plus-circle",
            layout=layout,
        )
        self.delete_node = widgets.Button(
            tooltip="Delete selected node (or double-click on the node)",
            icon="minus-circle",
            layout=layout,
        )
        self.create_script = widgets.Button(
            tooltip="Create script (or click the '+' tab)",
            icon="plus-square-o",
            layout=layout,
        )
        self.rename_script = widgets.Button(
            tooltip="Rename script",
            icon="pencil-square-o",  # Todo: Use file-pen once this is available
            layout=layout,
        )
        self.delete_script = widgets.Button(
            tooltip="Delete script",
            icon="minus-square-o",  # Todo: Use file-circle-minus once this is available
            layout=layout,
        )
        self.zero_location = widgets.Button(
            tooltip="Recenter script canvas at the origin",
            icon="map-marker",  # Todo: Use location-dot once this is available
            layout=layout,
        )
        self.zoom_in = widgets.Button(
            tooltip="Zoom canvas in", icon="search-plus", layout=layout
        )
        self.zoom_out = widgets.Button(
            tooltip="Zoom canvas out", icon="search-minus", layout=layout
        )

    def __iter__(self):
        """Iterates like a list based on order of attribute declaration"""
        return self.__dict__.values().__iter__()


class Toolbar(Box):
    box_class = widgets.HBox

    def __init__(self):
        super().__init__()
        alg_modes = ["data", "exec"]
        self.alg_mode_dropdown = widgets.Dropdown(
            options=alg_modes,
            value=alg_modes[0],
            disabled=False,
            layout=widgets.Layout(width="80px"),
        )
        self.buttons = Buttons()
        self.box.children = [self.alg_mode_dropdown, *self.buttons]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Widgets with which to populate the `input_widget` field of a new `ironflow.model.node.Node`.
"""

from ironflow.gui.boxes.node_interface.input_widgets import SliderControl



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Top-level objects for getting the front and back end (and various parts of the front end) to talk to each other.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type

import ipywidgets as widgets
from IPython.display import HTML

from ironflow.gui.boxes import (
    Toolbar,
    NodeController,
    NodePresenter,
    TextOut,
    UserInput,
    FlowBox,
)
from ironflow.gui.canvas_widgets import FlowCanvas
from ironflow.model.model import HasSession

if TYPE_CHECKING:
    from ironflow.model.node import Node
    from ironflow.gui.canvas_widgets.nodes import NodeWidget

debug_view = widgets.Output(layout={"border": "1px solid black"})


class GUI(HasSession):
    """
    The main ironflow object, connecting a ryven backend with a jupyter-friendly ipywidgets+ipycanvas frontend.

    Methods:
        draw: Build the ipywidget to interact with.
        register_user_node: Register with ironflow a new node from the current python process.
    """

    def __init__(
        self,
        session_title: str,
        extra_nodes_packages: Optional[list] = None,
        script_title: Optional[str] = None,
    ):
        """
        Create a new gui instance.

        Args:
            session_title (str): Title of the session to use. Will look for a json file of the same name and try to
                read it. If no such file exists, simply makes a new script instead.
            extra_nodes_packages (list | None): an optional list of nodes to register at instantiation. List items can
                be either a list of `ironflow.model.node.Node` subclasses, a module containing such subclasses, or a .py
                file of a module containing such subclasses. In all cases only those subclasses with the name pattern
                `*_Node` will be registered. (Default is None, don't register any extra nodes.)
            script_title (str|None): Title for an initial script. (Default is None, which generates "script_0" if a
                new script is needed on initialization, i.e. when existing session data cannot be read.)
        """
        super().__init__(
            session_title=session_title, extra_nodes_packages=extra_nodes_packages
        )

        self.flow_canvases = []
        self.toolbar = Toolbar()
        self.node_controller = NodeController(self)
        self.node_presenter = NodePresenter()
        self.text_out = TextOut()
        self.input = UserInput()
        self.flow_box = FlowBox(self.nodes_dictionary)

        try:
            self.load(f"{self.session_title}.json")
            print(f"Loaded session data for {self.session_title}")
        except FileNotFoundError:
            print(
                f"No session data found for {self.session_title}, making a new script."
            )
            self.create_script(script_title)
        self.update_tabs()

    def create_script(
        self,
        title: Optional[str] = None,
        create_default_logs: bool = True,
        data: Optional[dict] = None,
    ) -> None:
        super().create_script(
            title=title, create_default_logs=create_default_logs, data=data
        )
        self.flow_canvases.append(FlowCanvas(gui=self))

    def delete_script(self) -> None:
        self.flow_canvases.pop(self.active_script_index)
        self.node_controller.close()
        self.node_presenter.close()
        super().delete_script()

    @property
    def flow_canvas(self):
        return self.flow_canvases[self.active_script_index]

    @property
    def new_node_class(self):
        return self.flow_box.node_selector.new_node_class

    def serialize(self) -> dict:
        data = super().serialize()
        currently_active = self.active_script_index
        for i_script, script in enumerate(self.session.scripts):
            all_data = data["scripts"][i_script]["flow"]["nodes"]
            self.active_script_index = i_script
            for i, node_widget in enumerate(self.flow_canvas.objects_to_draw):
                all_data[i]["pos x"] = node_widget.x
                all_data[i]["pos y"] = node_widget.y
        self.active_script_index = currently_active
        return data

    def load_from_data(self, data: dict) -> None:
        super().load_from_data(data)
        self.flow_canvases = []
        for i_script, script in enumerate(self.session.scripts):
            flow_canvas = FlowCanvas(gui=self, flow=script.flow)
            all_data = data["scripts"][i_script]["flow"]["nodes"]
            for i_node, node in enumerate(script.flow.nodes):
                flow_canvas.load_node(
                    all_data[i_node]["pos x"], all_data[i_node]["pos y"], node
                )
            flow_canvas._built_object_to_gui_dict()
            flow_canvas.redraw()
            self.flow_canvases.append(flow_canvas)

    def register_node(self, node_class: Type[Node], node_group: Optional[str] = None):
        # Inherited __doc__ still applies just fine, all we do here is update a menu item afterwards.
        super().register_node(node_class=node_class, node_group=node_group)
        try:
            self.flow_box.node_selector.update(self.nodes_dictionary)
        except AttributeError:
            pass  # It's not defined yet in the super().__init__ call, which is fine

    def update_tabs(self):
        self.flow_box.update_tabs(
            outputs=[fc.output for fc in self.flow_canvases],
            titles=[fc.title for fc in self.flow_canvases],
            active_index=self.active_script_index,
        )
        for fc in self.flow_canvases:
            fc.display()

    def open_node_control(self, node: Node) -> None:
        self.node_controller.draw_for_node(node)

    def update_node_control(self) -> None:
        self.node_controller.draw()

    def close_node_control(self) -> None:
        self.node_controller.node = None
        self.node_controller.clear_output()

    def ensure_node_not_controlled(self, node: Node) -> None:
        if self.node_controller.node == node:
            self.node_controller.draw_for_node(None)

    def open_node_presenter(self, node_widget: NodeWidget):
        self.node_presenter.node_widget = node_widget

    def update_node_presenter(self):
        self.node_presenter.draw()

    def close_node_presenter(self):
        self.node_presenter.close()

    def ensure_node_not_presented(self, node_widget: NodeWidget) -> None:
        if self.node_presenter.node_widget == node_widget:
            self.node_presenter.node_widget = None

    def redraw_active_flow_canvas(self):
        self.flow_canvas.redraw()

    def print(self, msg: str):
        self.text_out.print(msg)

    @debug_view.capture(clear_output=True)
    def draw(self) -> widgets.VBox:
        """
        Build the gui.

        Returns:
            ipywidgets.VBox: The gui.
        """

        # Wire callbacks
        self.toolbar.alg_mode_dropdown.observe(
            self._change_alg_mode_dropdown, names="value"
        )
        self.toolbar.buttons.help_node.on_click(self._click_node_help)
        self.toolbar.buttons.load.on_click(self._click_load)
        self.toolbar.buttons.save.on_click(self._click_save)
        self.toolbar.buttons.add_node.on_click(self._click_add_node)
        self.toolbar.buttons.delete_node.on_click(self._click_delete_node)
        self.toolbar.buttons.create_script.on_click(self._click_create_script)
        self.toolbar.buttons.rename_script.on_click(self._click_rename_script)
        self.toolbar.buttons.delete_script.on_click(self._click_delete_script)
        self.toolbar.buttons.zero_location.on_click(self._click_zero_location)
        self.toolbar.buttons.zoom_in.on_click(self._click_zoom_in)
        self.toolbar.buttons.zoom_out.on_click(self._click_zoom_out)
        self.flow_box.script_tabs.observe(self._change_script_tabs)

        return widgets.VBox(
            [
                self.toolbar.box,
                self.input.box,
                self.flow_box.box,
                self.text_out.box,
                widgets.HBox([self.node_controller.box, self.node_presenter.box]),
                debug_view,
            ]
        )

    # Type hinting for unused `change` argument in callbacks taken from ipywidgets docs:
    # https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20Events.html#Traitlet-events
    def _change_alg_mode_dropdown(self, change: dict) -> None:
        # Current behaviour: Updates the flow mode for all scripts
        # Todo: Change only for the active script, and update the dropdown on tab (script) switching
        for script in self.session.scripts:
            script.flow.set_algorithm_mode(self.toolbar.alg_mode_dropdown.value)

    def _click_save(self, change: dict) -> None:
        self.input.open_text(
            "Save file",
            self._click_confirm_save,
            self.session_title,
            description_tooltip="Save to file name (omit the file extension, .json)",
        )
        self.print("Choose a file name to save to (omit the file extension, .json)")

    def _click_confirm_save(self, change: dict) -> None:
        file_name = self.input.text
        self.save(f"{file_name}.json")
        self.print(f"Session saved to {file_name}.json")
        self.input.clear()

    def _click_load(self, change: dict) -> None:
        self.input.open_text(
            "Load file",
            self._click_confirm_load,
            self.session_title,
            description_tooltip="Load from file name (omit the file extension, .json).",
        )
        self.print("Choose a file name to load (omit the file extension, .json)")

    def _click_confirm_load(self, change: dict) -> None:
        file_name = self.input.text
        self.load(f"{file_name}.json")
        self.update_tabs()
        self.node_presenter.clear_output()
        self.print(f"Session loaded from {file_name}.json")
        self.input.clear()

    def _click_node_help(self, change: dict) -> None:
        def _pretty_docstring(node_class):
            """
            If we just pass a string, `display` doesn't resolve newlines.
            If we pass a `print`ed string, `display` also shows the `None` value returned by `print`
            So we use this ugly hack.
            """
            string = (
                f"{node_class.__name__.replace('_Node', '')}:\n{node_class.__doc__}"
            )
            return HTML(
                string.replace("\n", "<br>")
                .replace("\t", "&emsp;")
                .replace(" ", "&nbsp;")
            )

        self.print(_pretty_docstring(self.new_node_class))

    def _click_add_node(self, change: dict) -> None:
        self.flow_canvas.add_node(10, 10, self.new_node_class)

    def _click_delete_node(self, change: dict) -> None:
        self.flow_canvas.delete_selected()

    def _click_create_script(self, change: dict) -> None:
        self.create_script()
        self.update_tabs()

    def _click_rename_script(self, change: dict) -> None:
        self.input.open_text(
            "New name",
            self._click_confirm_rename,
            self.script.title,
            description_tooltip="New script name",
        )
        self.print("Choose a new name for the current script")

    def _click_confirm_rename(self, change: dict) -> None:
        new_name = self.input.text
        old_name = self.script.title
        rename_success = self.rename_script(new_name)
        if rename_success:
            self.flow_box.script_tabs.set_title(self.active_script_index, new_name)
            self.print(f"Script '{old_name}' renamed '{new_name}'")
        else:
            self.print(
                f"INVALID NAME: Failed to rename script '{self.script.title}' to '{new_name}'."
            )

    def _click_delete_script(self, change: dict) -> None:
        self.input.open_bool(
            f"Delete the entire script {self.script.title}?",
            self._click_confirm_delete_script,
        )

    def _click_confirm_delete_script(self, change: dict) -> None:
        script_name = self.script.title
        self.delete_script()
        self.update_tabs()
        self.print(f"Script {script_name} deleted")

    def _click_zero_location(self, change: dict) -> None:
        self.flow_canvas.x = 0
        self.flow_canvas.y = 0
        self.flow_canvas.redraw()

    def _click_zoom_in(self, change: dict) -> None:
        self.flow_canvas.zoom_in()

    def _click_zoom_out(self, change: dict) -> None:
        self.flow_canvas.zoom_out()

    def _click_input_text_cancel(self, change: dict) -> None:
        self.input.clear()
        self.text_out.clear()

    def _change_script_tabs(self, change: dict):
        if change["name"] == "selected_index" and change["new"] is not None:
            self.input.clear()
            self.flow_canvas.deselect_all()
            if self.flow_box.script_tabs.selected_index == self.n_scripts:
                self.create_script()
                self.update_tabs()
            else:
                self.active_script_index = self.flow_box.script_tabs.selected_index



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Canvas widgets for ryven IO ports.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from ironflow.gui.canvas_widgets.base import HideableWidget, CanvasWidget
from ironflow.gui.canvas_widgets.layouts import PortLayout

if TYPE_CHECKING:
    from ironflow.gui.canvas_widgets.flow import FlowCanvas
    from ironflow.gui.canvas_widgets.base import Number
    from ironflow.model import NodePort


class PortWidget(HideableWidget):
    def __init__(
        self,
        x: Number,
        y: Number,
        parent: FlowCanvas | CanvasWidget,
        layout: PortLayout,
        port: NodePort,
        selected: bool = False,
        title: Optional[str] = None,
        hidden_x: Optional[Number] = None,
        hidden_y: Optional[Number] = None,
        radius: Number = 10,
    ):
        super().__init__(
            x=x,
            y=y,
            parent=parent,
            layout=layout,
            selected=selected,
            title=title if title is not None else port.label_str,
            hidden_x=hidden_x,
            hidden_y=hidden_y,
        )

        self.radius = radius
        self.port = port

    def on_click(
        self, last_selected_object: Optional[CanvasWidget]
    ) -> PortWidget | None:
        if last_selected_object == self:
            self.deselect()
            return None
        elif isinstance(last_selected_object, PortWidget):
            self.flow.connect_nodes(last_selected_object.port, self.port)
            self.deselect_all()
            return None
        else:
            if last_selected_object is not None:
                last_selected_object.deselect()
            self.select()
            return self

    def draw_shape(self) -> None:
        self.canvas.fill_style = (
            self.layout.selected_color
            if self.selected
            else self.layout.background_color
        )
        self.canvas.fill_circle(self.x, self.y, self.radius)

    def draw_title(self) -> None:
        self.canvas.font = self.layout.font_string
        self.canvas.fill_style = self.layout.font_color
        self.canvas.fill_text(
            self.title, self.x + self.radius + 3, self.y + self.radius // 2
        )

    def _is_at_xy(self, x_in: Number, y_in: Number) -> bool:
        return (x_in - self.x) ** 2 + (y_in - self.y) ** 2 < self.radius**2



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Specialty widgets for directly controlling node input.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import ipywidgets as widgets

if TYPE_CHECKING:
    from ironflow.model.node import Node
    from ironflow.gui.gui import GUI


class SliderControl:
    def __init__(self, gui: GUI, node: Node):
        self.gui = gui
        self.node = node
        self.widget = widgets.FloatSlider(
            value=self.node.val, min=0, max=10, continuous_update=False
        )

        self.widget.observe(self.widget_change, names="value")

    def widget_change(self, change: dict) -> None:
        self.node.set_state({"val": change["new"]}, 0)
        self.node.update_event()
        self.gui.redraw_active_flow_canvas()



"""
Setuptools based setup module
"""
from setuptools import setup, find_packages
import versioneer

setup(
    name='ironflow',
    version=versioneer.get_version(),
    description='ironflow - A visual scripting interface for pyiron.',
    long_description='Ironflow combines ryven, ipycanvas and ipywidgets to provide a Jupyter-based visual scripting '
                     'gui for running pyiron workflow graphs.',

    url='https://github.com/pyiron/ironflow',
    author='Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department',
    author_email='liamhuber@greyhavensolutions.com',
    license='BSD',

    classifiers=['Development Status :: 5 - Production/Stable',
                 'Topic :: Scientific/Engineering :: Physics',
                 'License :: OSI Approved :: BSD License',
                 'Intended Audience :: Science/Research',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: 3.9'],

    keywords='pyiron',
    packages=find_packages(exclude=["*tests*", "*docs*", "*binder*", "*conda*", "*notebooks*", "*.ci_support*"]),
    install_requires=[
        'ipycanvas',
        'ipython',
        'ipywidgets >= 7,< 8',
        'matplotlib',
        'nglview',
        'numpy',
        'pyiron_base',
        'pyiron_atomistics',
        'ryvencore'
    ],
    cmdclass=versioneer.get_cmdclass(),

    )



import code
from contextlib import redirect_stdout, redirect_stderr

from ironflow.model import dtypes, NodeInputBP, NodeOutputBP
from ironflow.model.node import Node, PlaceholderWidgetsContainer
from ironflow.gui.boxes.node_interface.input_widgets import SliderControl

widgets = PlaceholderWidgetsContainer()


class NodeBase(Node):
    version = "v0.1"
    color = "#FFCA00"


class DualNodeBase(NodeBase):
    """For nodes that can be active and passive"""

    version = "v0.1"

    def __init__(self, params, active=True):
        super().__init__(params)

        self.active = active
        if active:
            self.actions["make passive"] = {"method": self.make_passive}
        else:
            self.actions["make active"] = {"method": self.make_active}

    def make_passive(self):
        del self.actions["make passive"]

        self.delete_input(0)
        self.delete_output(0)
        self.active = False

        self.actions["make active"] = {"method": self.make_active}

    def make_active(self):
        del self.actions["make active"]

        self.create_input(type_="exec", insert=0)
        self.create_output(type_="exec", insert=0)
        self.active = True

        self.actions["make passive"] = {"method": self.make_passive}

    def get_state(self) -> dict:
        return {"active": self.active}

    def set_state(self, data: dict, version):
        self.active = data["active"]


# -------------------------------------------


class Checkpoint_Node(NodeBase):
    """Provides a simple checkpoint to reroute your connections"""

    title = "checkpoint"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="data"),
    ]
    init_outputs = [
        NodeOutputBP(type_="data"),
    ]
    style = "small"

    def __init__(self, params):
        super().__init__(params)

        self.display_title = ""

        self.active = False

        # initial actions
        self.actions["add output"] = {"method": self.add_output}
        self.actions["remove output"] = {"0": {"method": self.remove_output, "data": 0}}
        self.actions["make active"] = {"method": self.make_active}

    """State transitions"""

    def clear_ports(self):
        # remove all outputs
        for i in range(len(self.outputs)):
            self.delete_output(0)

        # remove all inputs
        for i in range(len(self.inputs)):
            self.delete_input(0)

    def make_active(self):
        self.active = True

        # rebuild inputs and outputs
        self.clear_ports()
        self.create_input(type_="exec")
        self.create_output(type_="exec")

        # update actions
        del self.actions["make active"]
        self.actions["make passive"] = {"method": self.make_passive}
        self.actions["remove output"] = {"0": {"method": self.remove_output, "data": 0}}

    def make_passive(self):
        self.active = False

        # rebuild inputs and outputs
        self.clear_ports()
        self.create_input(type_="data")
        self.create_output(type_="data")

        # update actions
        del self.actions["make passive"]
        self.actions["make active"] = {"method": self.make_active}
        self.actions["remove output"] = {"0": {"method": self.remove_output, "data": 0}}

    """Actions"""

    def add_output(self):
        index = len(self.outputs)

        if self.active:
            self.create_output(type_="exec")
        else:
            self.create_output(type_="data")

        self.actions["remove output"][str(index)] = {
            "method": self.remove_output,
            "data": index,
        }

    def remove_output(self, index):
        self.delete_output(index)

        del self.actions["remove output"][str(len(self.outputs))]

    """Behavior"""

    def update_event(self, inp=-1):
        if self.active and inp == 0:
            for i in range(len(self.outputs)):
                self.exec_output(i)

        elif not self.active:
            data = self.input(0)
            for i in range(len(self.outputs)):
                self.set_output_val(i, data)

    """State Reload"""

    def get_state(self) -> dict:
        return {
            "active": self.active,
            "num outputs": len(self.outputs),
        }

    def set_state(self, data: dict, version):
        self.actions["remove output"] = {
            {"method": self.remove_output, "data": i}
            for i in range(data["num outputs"])
        }

        if data["active"]:
            self.make_active()


class Button_Node(NodeBase):
    title = "Button"
    version = "v0.1"
    main_widget_class = widgets.ButtonNode_MainWidget
    main_widget_pos = "between ports"
    init_inputs = []
    init_outputs = [NodeOutputBP(type_="exec")]
    color = "#99dd55"

    def update_event(self, inp=-1):
        self.exec_output(0)


class Print_Node(DualNodeBase):
    title = "Print"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec"),
        NodeInputBP(dtype=dtypes.Data(size="m")),
    ]
    init_outputs = [
        NodeOutputBP(type_="exec"),
    ]
    color = "#5d95de"

    def __init__(self, params):
        super().__init__(params, active=True)

    def update_event(self, inp=-1):
        if self.active and inp == 0:
            self.val = self.input(1)
        elif not self.active:
            self.val = self.input(0)


import logging


class Log_Node(DualNodeBase):
    title = "Log"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(type_="exec"),
        NodeInputBP("msg", type_="data"),
    ]
    init_outputs = [
        NodeOutputBP(type_="exec"),
    ]
    main_widget_class = widgets.LogNode_MainWidget
    main_widget_pos = "below ports"
    color = "#5d95de"

    def __init__(self, params):
        super().__init__(params, active=True)

        self.logger = self.new_logger("Log Node")

        self.targets = {
            **self.script.logs_manager.default_loggers,
            "own": self.logger,
        }
        self.target = "global"

    def update_event(self, inp=-1):
        if self.active and inp == 0:
            i = 1
        elif not self.active:
            i = 0
        else:
            return

        msg = self.input(i)

        self.targets[self.target].log(logging.INFO, msg=msg)

    def get_state(self) -> dict:
        return {
            **super().get_state(),
            "target": self.target,
        }

    def set_state(self, data: dict, version):
        super().set_state(data, version)
        self.target = data["target"]
        if self.session.gui and self.main_widget():
            self.main_widget().set_target(self.target)


class Clock_Node(NodeBase):
    title = "clock"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Float(default=0.1), label="delay"),
        NodeInputBP(
            dtype=dtypes.Integer(default=-1, bounds=(-1, 1000)), label="iterations"
        ),
    ]
    init_outputs = [NodeOutputBP(type_="exec")]
    color = "#5d95de"
    main_widget_class = widgets.ClockNode_MainWidget
    main_widget_pos = "below ports"

    def __init__(self, params):
        super().__init__(params)

        self.actions["start"] = {"method": self.start}
        self.actions["stop"] = {"method": self.stop}

        if self.session.gui:

            from qtpy.QtCore import QTimer

            self.timer = QTimer(self)
            self.timer.timeout.connect(self.timeouted)
            self.iteration = 0

    def timeouted(self):
        self.exec_output(0)
        self.iteration += 1
        if -1 < self.input(1) <= self.iteration:
            self.stop()

    def start(self):
        if self.session.gui:
            self.timer.setInterval(self.input(0) * 1000)
            self.timer.start()
        else:
            import time

            for i in range(self.input(1)):
                self.exec_output(0)
                time.sleep(self.input(0))

    def stop(self):
        self.iteration = 0
        if self.session.gui:
            self.timer.stop()

    def toggle(self):
        # triggered from main widget
        if self.session.gui:
            if self.timer.isActive():
                self.stop()
            else:
                self.start()

    def update_event(self, inp=-1):
        if self.session.gui:
            self.timer.setInterval(self.input(0) * 1000)

    def remove_event(self):
        self.stop()


class Slider_Node(NodeBase):
    title = "slider"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(dtype=dtypes.Integer(default=1), label="scl"),
        NodeInputBP(dtype=dtypes.Boolean(default=False), label="round"),
    ]
    init_outputs = [
        NodeOutputBP(),
    ]
    main_widget_class = widgets.SliderNode_MainWidget
    main_widget_pos = "below ports"
    input_widget = SliderControl

    def __init__(self, params):
        super().__init__(params)

        self.val = 0

    def place_event(self):
        self.update()

    def view_place_event(self):
        # when running in gui mode, the value might come from the input widget
        self.update()

    def update_event(self, inp=-1):

        v = self.input(0) * self.val
        if self.input(1):
            v = round(v)

        self.set_output_val(0, v)

    def get_state(self) -> dict:
        return {
            "val": self.val,
        }

    def set_state(self, data: dict, version):
        self.val = data["val"]


class _DynamicPorts_Node(NodeBase):
    version = "v0.1"
    init_inputs = []
    init_outputs = []

    def __init__(self, params):
        super().__init__(params)

        self.actions["add input"] = {"method": self.add_inp}
        self.actions["add output"] = {"method": self.add_out}

        self.num_inputs = 0
        self.num_outputs = 0

    def add_inp(self):
        self.create_input()

        index = self.num_inputs
        self.actions[f"remove input {index}"] = {
            "method": self.remove_inp,
            "data": index,
        }

        self.num_inputs += 1

    def remove_inp(self, index):
        self.delete_input(index)
        self.num_inputs -= 1
        del self.actions[f"remove input {self.num_inputs}"]

    def add_out(self):
        self.create_output()

        index = self.num_outputs
        self.actions[f"remove output {index}"] = {
            "method": self.remove_out,
            "data": index,
        }

        self.num_outputs += 1

    def remove_out(self, index):
        self.delete_output(index)
        self.num_outputs -= 1
        del self.actions[f"remove output {self.num_outputs}"]

    def get_state(self) -> dict:
        return {
            "num inputs": self.num_inputs,
            "num outputs": self.num_outputs,
        }

    def set_state(self, data: dict):
        self.num_inputs = data["num inputs"]
        self.num_outputs = data["num outputs"]


class Exec_Node(_DynamicPorts_Node):
    title = "exec"
    version = "v0.1"
    main_widget_class = widgets.CodeNode_MainWidget
    main_widget_pos = "between ports"

    def __init__(self, params):
        super().__init__(params)

        self.code = None

    def place_event(self):
        pass

    def update_event(self, inp=-1):
        exec(self.code)

    def get_state(self) -> dict:
        return {
            **super().get_state(),
            "code": self.code,
        }

    def set_state(self, data: dict, version):
        super().set_state(data, version)
        self.code = data["code"]


class Eval_Node(NodeBase):
    title = "eval"
    version = "v0.1"
    init_inputs = [
        # NodeInputBP(),
    ]
    init_outputs = [
        NodeOutputBP(),
    ]
    main_widget_class = widgets.EvalNode_MainWidget
    main_widget_pos = "between ports"

    def __init__(self, params):
        super().__init__(params)

        self.actions["add input"] = {"method": self.add_param_input}

        self.number_param_inputs = 0
        self.expression_code = None

    def place_event(self):
        if self.number_param_inputs == 0:
            self.add_param_input()

    def add_param_input(self):
        self.create_input()

        index = self.number_param_inputs
        self.actions[f"remove input {index}"] = {
            "method": self.remove_param_input,
            "data": index,
        }

        self.number_param_inputs += 1

    def remove_param_input(self, index):
        self.delete_input(index)
        self.number_param_inputs -= 1
        del self.actions[f"remove input {self.number_param_inputs}"]

    def update_event(self, inp=-1):
        inp = [self.input(i) for i in range(self.number_param_inputs)]
        self.set_output_val(0, eval(self.expression_code))

    def get_state(self) -> dict:
        return {
            "num param inputs": self.number_param_inputs,
            "expression code": self.expression_code,
        }

    def set_state(self, data: dict, version):
        self.number_param_inputs = data["num param inputs"]
        self.expression_code = data["expression code"]


class Interpreter_Node(NodeBase):
    """Provides a python interpreter via a basic console with access to the
    node's properties."""

    title = "interpreter"
    version = "v0.1"
    init_inputs = []
    init_outputs = []
    main_widget_class = widgets.InterpreterConsole

    # DEFAULT COMMANDS

    def clear(self):
        self.hist.clear()
        self._hist_updated()

    def reset(self):
        self.interp = code.InteractiveInterpreter(locals=locals())

    COMMANDS = {
        "clear": clear,
        "reset": reset,
    }

    def __init__(self, params):
        super().__init__(params)

        self.interp = None
        self.hist: [str] = []
        self.buffer: [str] = []

        self.reset()

    def _hist_updated(self):
        if self.session.gui:
            self.main_widget().interp_updated()

    def process_input(self, cmds: str):
        m = self.COMMANDS.get(cmds)
        if m is not None:
            m()
        else:
            for l in cmds.splitlines():
                self.write(l)  # print input
                self.buffer.append(l)
            src = "\n".join(self.buffer)

            def run_src():
                more_inp_required = self.interp.runsource(src, "<console>")
                if not more_inp_required:
                    self.buffer.clear()

            if self.session.gui:
                with redirect_stdout(self), redirect_stderr(self):
                    run_src()
            else:
                run_src()

    def write(self, line: str):
        self.hist.append(line)
        self._hist_updated()


class Storage_Node(NodeBase):
    """Sequentially stores all the data provided at the input in an array.
    A COPY of the storage array is provided at the output"""

    title = "store"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(),
    ]
    init_outputs = [
        NodeOutputBP(),
    ]
    color = "#aadd55"

    def __init__(self, params):
        super().__init__(params)

        self.storage = []

        self.actions["clear"] = {"method": self.clear}

    def clear(self):
        self.storage.clear()
        self.set_output_val(0, [])

    def update_event(self, inp=-1):
        self.storage.append(self.input(0))
        self.set_output_val(0, self.storage.copy())

    def get_state(self) -> dict:
        return {
            "data": self.storage,
        }

    def set_state(self, data: dict, version):
        self.storage = data["data"]


import uuid


class LinkIN_Node(NodeBase):
    """You can use link OUT nodes to link them up to this node.
    Whenever a link IN node receives data (or an execution signal),
    if there is a linked OUT node, it will receive the data
    and propagate it further."""

    title = "link IN"
    version = "v0.1"
    init_inputs = [
        NodeInputBP(),
    ]
    init_outputs = []  # no outputs

    # instances registration
    INSTANCES = {}  # {UUID: node}

    def __init__(self, params):
        super().__init__(params)
        self.display_title = "link"

        # register
        self.ID: uuid.UUID = uuid.uuid4()
        self.INSTANCES[str(self.ID)] = self

        self.actions["add input"] = {"method": self.add_inp}
        self.actions["remove inp"] = {}
        self.actions["copy ID"] = {"method": self.copy_ID}

        self.linked_node: LinkOUT_Node = None

    def copy_ID(self):
        from qtpy.QtWidgets import QApplication

        QApplication.clipboard().setText(str(self.ID))

    def add_inp(self):
        index = len(self.inputs)

        self.create_input()

        self.actions["remove inp"][str(index)] = {
            "method": self.rem_inp,
            "data": index,
        }
        if self.linked_node is not None:
            self.linked_node.add_out()

    def rem_inp(self, index):
        self.delete_input(index)
        del self.actions["remove inp"][str(len(self.inputs))]
        if self.linked_node is not None:
            self.linked_node.rem_out(index)

    def update_event(self, inp=-1):
        if self.linked_node is not None:
            self.linked_node.set_output_val(inp, self.input(inp))

    def get_state(self) -> dict:
        return {
            "ID": str(self.ID),
        }

    def set_state(self, data: dict, version):
        if data["ID"] in self.INSTANCES:
            # this happens when some existing node has been copied and pasted.
            # we only want to rebuild links when loading a project, considering
            # new links when copying nodes might get quite complex
            pass
        else:
            del self.INSTANCES[str(self.ID)]  # remove old ref
            self.ID = uuid.UUID(data["ID"])  # use original ID
            self.INSTANCES[str(self.ID)] = self  # set new ref

            # resolve possible pending link builds from OUT nodes that happened
            # to get initialized earlier
            LinkOUT_Node.new_link_in_loaded(self)

    def remove_event(self):
        # break existent link
        if self.linked_node:
            self.linked_node.linked_node = None
            self.linked_node = None


class LinkOUT_Node(NodeBase):
    """The complement to the link IN node"""

    title = "link OUT"
    version = "v0.1"
    init_inputs = []  # no inputs
    init_outputs = []  # will be synchronized with linked IN node

    INSTANCES = []
    PENDING_LINK_BUILDS = {}
    # because a link OUT node might get initialized BEFORE it's corresponding
    # link IN, it then stores itself together with the ID of the link IN it's
    # waiting for in PENDING_LINK_BUILDS

    @classmethod
    def new_link_in_loaded(cls, n: LinkIN_Node):
        for out_node, in_ID in cls.PENDING_LINK_BUILDS.items():
            if in_ID == str(n.ID):
                out_node.link_to(n)

    def __init__(self, params):
        super().__init__(params)
        self.display_title = "link"

        self.INSTANCES.append(self)
        self.linked_node: LinkIN_Node = None

        self.actions["link to ID"] = {"method": self.choose_link_ID}

    def choose_link_ID(self):
        """opens a small input dialog for providing a copied link IN ID"""

        from qtpy.QtWidgets import QDialog, QMessageBox, QVBoxLayout, QLineEdit

        class IDInpDialog(QDialog):
            def __init__(self):
                super().__init__()
                self.id_str = None
                self.setLayout(QVBoxLayout())
                self.line_edit = QLineEdit()
                self.layout().addWidget(self.line_edit)
                self.line_edit.returnPressed.connect(self.return_pressed)

            def return_pressed(self):
                self.id_str = self.line_edit.text()
                self.accept()

        d = IDInpDialog()
        d.exec_()

        if d.id_str is not None:
            n = LinkIN_Node.INSTANCES.get(d.id_str)
            if n is None:
                QMessageBox.warning(
                    title="link failed", text="couldn't find a valid link in node"
                )
            else:
                self.link_to(n)

    def link_to(self, n: LinkIN_Node):
        self.linked_node = n
        n.linked_node = self

        o = len(self.outputs)
        i = len(self.linked_node.inputs)

        # remove outputs if there are too many
        for j in range(i, o):
            self.delete_output(0)

        # add outputs if there are too few
        for j in range(o, i):
            self.create_output()

        self.update()

    def add_out(self):
        # triggered by linked_node
        self.create_output()

    def rem_out(self, index):
        # triggered by linked_node
        self.delete_output(index)

    def update_event(self, inp=-1):
        if self.linked_node is None:
            return

        # update ALL ports
        for i in range(len(self.outputs)):
            self.set_output_val(i, self.linked_node.input(i))

    def get_state(self) -> dict:
        if self.linked_node is None:
            return {}
        else:
            return {
                "linked ID": str(self.linked_node.ID),
            }

    def set_state(self, data: dict, version):
        if len(data) > 0:
            n: LinkIN_Node = LinkIN_Node.INSTANCES.get(data["linked ID"])
            if n is None:
                # means that the OUT node gets initialized before it's link IN
                self.PENDING_LINK_BUILDS[self] = data["linked ID"]
            elif n.linked_node is None:
                # pair up
                n.linked_node = self
                self.linked_node = n

    def remove_event(self):
        # break existent link
        if self.linked_node:
            self.linked_node.linked_node = None
            self.linked_node = None


# -------------------------------------------


nodes = [
    Checkpoint_Node,
    Button_Node,
    Print_Node,
    Log_Node,
    Clock_Node,
    Slider_Node,
    Exec_Node,
    Eval_Node,
    Storage_Node,
    LinkIN_Node,
    LinkOUT_Node,
    Interpreter_Node,
]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
Widgets for interacting with the flow diagram.
"""

from __future__ import annotations

import ipywidgets as widgets

from ironflow.gui.boxes.base import Box


class NodeSelector(Box):
    box_class = widgets.VBox

    def __init__(self, nodes_dictionary):
        super().__init__()
        self._nodes_dictionary = nodes_dictionary

        self.modules_dropdown = widgets.Dropdown(
            options=self.module_options,
            value=list(self.module_options)[0],
            disabled=False,
            layout=widgets.Layout(width="130px"),
        )

        self.node_selector = widgets.RadioButtons(
            options=self.nodes_options,
            value=list(self.nodes_options)[0],
            disabled=False,
        )

        self.modules_dropdown.observe(self.change_modules_dropdown, names="value")

        self.box.children = [self.modules_dropdown, self.node_selector]

    def change_modules_dropdown(self, change: dict) -> None:
        self.node_selector.options = sorted(
            self._nodes_dictionary[self.modules_dropdown.value].keys()
        )

    @property
    def new_node_class(self):
        return self._nodes_dictionary[self.modules_dropdown.value][
            self.node_selector.value
        ]

    @property
    def module_options(self) -> list[str]:
        return sorted(self._nodes_dictionary.keys())

    @property
    def nodes_options(self) -> list[str]:
        return sorted(self._nodes_dictionary[self.modules_dropdown.value].keys())

    def update(self, nodes_dictionary: dict) -> None:
        self._nodes_dictionary = nodes_dictionary
        self.modules_dropdown.options = self.module_options


class FlowBox(Box):
    box_class = widgets.HBox

    def __init__(self, nodes_dictionary: dict):
        super().__init__()
        self.node_selector = NodeSelector(nodes_dictionary=nodes_dictionary)
        self.script_tabs = widgets.Tab([])

        self.node_selector.box.layout.width = "15%"
        self.script_tabs.layout.width = "85%"

        self.box.children = [self.node_selector.box, self.script_tabs]

    def update_nodes(self, nodes_dictionary: dict):
        self.node_selector.update(nodes_dictionary=nodes_dictionary)

    def update_tabs(
        self, outputs: list[widgets.Output], titles: list[str], active_index: int
    ):
        self.script_tabs.selected_index = None
        # ^ To circumvent a bug where the index gets set to 0 on child changes
        # https://github.com/jupyter-widgets/ipywidgets/issues/2988
        self.script_tabs.children = outputs
        for i, title in enumerate(titles):
            self.script_tabs.set_title(i, title)
        self._add_new_script_tab()
        self.script_tabs.selected_index = active_index

    def _add_new_script_tab(self):
        self.script_tabs.children += (
            widgets.Output(layout={"border": "1px solid black"}),
        )
        self.script_tabs.set_title(len(self.script_tabs.children) - 1, "+")



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""
The back-end model which interfaces with Ryven.
"""

from __future__ import annotations

import importlib.util
import json
import types
from abc import ABC
from inspect import isclass
from pathlib import Path
from types import ModuleType
from typing import Optional, Type

from ryvencore import Session, Script, Flow

from ironflow.model.node import Node


class HasSession(ABC):
    """Mixin for an object which has a Ryven session as the underlying model"""

    def __init__(self, session_title: str, extra_nodes_packages: Optional[list] = None):
        self._session = Session()
        self.session_title = session_title
        self._active_script_index = 0

        self.nodes_dictionary = {}
        from ironflow.nodes import built_in
        from ironflow.nodes.pyiron import atomistics_nodes
        from ironflow.nodes.std import (
            basic_operators,
            control_structures,
            special_nodes,
        )

        for module in [
            built_in,
            atomistics_nodes,
            basic_operators,
            control_structures,
            special_nodes,
        ]:
            self.register_nodes_from_module(module)

        if extra_nodes_packages is not None:
            for package in extra_nodes_packages:
                # Todo: Figure out how to allow node_group to be passed in in an elegant way here for each package
                self.register_nodes(package)

    @property
    def active_script_index(self) -> int:
        return self._active_script_index

    @active_script_index.setter
    def active_script_index(self, i: int) -> None:
        if i >= len(self.session.scripts):
            raise KeyError(
                f"Attempted to activate script {i}, but there are only {len(self.session.scripts)} available."
            )
        self._active_script_index = i % self.n_scripts

    @property
    def session(self) -> Session:
        return self._session

    @property
    def script(self) -> Script:
        return self.session.scripts[self.active_script_index]

    @property
    def flow(self) -> Flow:
        return self.script.flow

    @property
    def n_scripts(self):
        return len(self.session.scripts)

    @property
    def next_auto_script_name(self):
        i = 0
        titles = [s.title for s in self.session.scripts]
        while f"script_{i}" in titles:
            i += 1
        return f"script_{i}"

    def create_script(
        self,
        title: Optional[str] = None,
        create_default_logs: bool = True,
        data: Optional[dict] = None,
    ) -> None:
        self.session.create_script(
            title=title if title is not None else self.next_auto_script_name,
            create_default_logs=create_default_logs,
            data=data,
        )
        self.active_script_index = self.n_scripts - 1

    def delete_script(self) -> None:
        last_active = self.active_script_index
        self.session.delete_script(self.script)
        if self.n_scripts == 0:
            self.create_script()
        else:
            self.active_script_index = last_active - 1

    def rename_script(self, new_name: str) -> bool:
        return self.session.rename_script(self.script, new_name)

    def save(self, file_path: str) -> None:
        data = self.serialize()

        with open(file_path, "w") as f:
            f.write(json.dumps(data, indent=4))

    def serialize(self) -> dict:
        return self.session.serialize()

    def load(self, file_path: str) -> None:
        with open(file_path, "r") as f:
            data = json.loads(f.read())

        self.load_from_data(data)

    def load_from_data(self, data: dict) -> None:
        for script in self.session.scripts[::-1]:
            self.session.delete_script(script)
        self.session.load(data)
        self.active_script_index = 0

    def register_node(
        self, node_class: Type[Node], node_group: Optional[str] = None
    ) -> None:
        """
        Registers a node class with the ryven session and model, storing it in `nodes_dictionary`. Some node attributes
        (`identifier_prefix` and `type_`) are also set on the Node class.

        Args:
            node_class (Type[Node]): The node class to register.
            node_group (str | None): The sub-collection to which this node belongs. (Default is None, which uses the
                last bit of the module path.)


        Note: The sub-collection in the `nodes_dictionary` to which the node gets added depends only on the *tail* of its
              module path, so it is possible that nodes from two different sources get grouped together. In case this
              leads to a conflict, `node_module` can be explicitly provided and this will be used instead.

        Note: You can re-register a class to update its functionality, but only *newly placed* nodes will see this
                update. Already-placed nodes are still instances of the old class and need to be deleted.

        Note: You can save the graph as normal, but new gui instances will need to register the same custom nodes before
            loading the saved graph is possible.

        Example:
            >>> from ironflow import GUI
            >>> from ironflow.custom_nodes import Node, NodeInputBP, NodeOutputBP, dtypes, input_widgets
            >>> gui = GUI(script_title='foo')
            >>>
            >>> class MyNode(Node):
            >>>     title = "MyUserNode"
            >>>     init_inputs = [
            >>>         NodeInputBP(dtype=dtypes.Integer(default=1), label="foo")
            >>>     ]
            >>>     init_outputs = [
            >>>        NodeOutputBP(label="bar")
            >>>    ]
            >>>    color = 'cyan'
            >>>
            >>>     def update_event(self, inp=-1):
            >>>         self.set_output_val(0, self.input(0) + 42)
            >>>
            >>> gui.register_node(MyNode)
        """
        if node_class in self.session.nodes:
            self.session.unregister_node(node_class)
        self.session.register_node(node_class)

        module = node_class.__module__
        identifier_prefix, _, module_shorthand = module.rpartition(".")
        node_class.identifier_prefix = (
            identifier_prefix if node_class.identifier is None else None
        )
        node_class.type_ = (
            module + node_class.type_ if not node_class.type_ else node_class.type_
        )

        node_group = node_group or module_shorthand
        if node_group not in self.nodes_dictionary.keys():
            self.nodes_dictionary[node_group] = {}
        self.nodes_dictionary[node_group][node_class.title] = node_class

    def register_nodes_from_module(
        self, module: ModuleType, node_group: Optional[str] = None
    ) -> None:
        """
        Search through the provided python module for all subclasses `ironflow.main.node.Node` whose name ends with
        `'_Node'` and register them with the ryven session and the model's `nodes_dictionary`.

        Args:
            module (types.ModuleType): The module to register from.
            node_group (str | None): The sub-collection to which this node belongs. (Default is None, which uses the
                last bit of the module path.)
        """
        for name in [key for key in module.__dir__() if key.endswith("_Node")]:
            node = getattr(module, name)
            if not isclass(node) or not issubclass(node, Node):
                raise TypeError(
                    f"Tried to import {name} from {module}, but it was a {node.__class__} instead of {Node}"
                )
            self.register_node(node_class=node, node_group=node_group)

    def register_nodes_from_file(
        self, file_path: str | Path, node_group: Optional[str] = None
    ) -> None:
        """
        Loads a .py file as a module, then searches through it for all subclasses `ironflow.main.node.Node` whose name
        ends with `'_Node'` and register them with the ryven session and the model's `nodes_dictionary`.

        Args:
            file_path (str | pathlib.Path): The .py file to load.
            node_group (str | None): The sub-collection to which this node belongs. (Default is None, which uses the
                file name stripped of its .py suffix.)
        """
        path = Path(file_path)
        resolved = path.resolve().__str__()
        if not path.is_file():
            raise ValueError(f"No file found at {resolved}")

        spec = importlib.util.spec_from_file_location(
            resolved.replace("/", ".").lstrip(".").rpartition(".")[0], resolved
        )
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        self.register_nodes_from_module(module, node_group=node_group)

    def register_nodes(
        self,
        source: str | Path | types.ModuleType | list | tuple,
        node_group: Optional[str] = None,
    ) -> None:
        if isinstance(source, (str, Path)):
            self.register_nodes_from_file(source, node_group=node_group)
        elif isinstance(source, types.ModuleType):
            self.register_nodes_from_module(source, node_group=node_group)
        elif isinstance(source, (list, tuple)) and all(
            [issubclass(item, Node) for item in source]
        ):
            for node_class in source:
                self.register_node(node_class, node_group=node_group)
